# -*- coding: utf-8 -*-
# Morra project: Morphological parser 3
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Get the results of join and separate FEATS-2 taggers and make adjusted
tagging on the ground of them. POS tagger did not change.
"""
from copy import deepcopy

from morra.morph_parser2 import MorphParser2


class MorphParser3(MorphParser2):

    def predict_feats3(self, sentence,
                       with_s_backoff=True, max_s_repeats=0,
                       with_j_backoff=True, max_j_repeats=0,
                       inplace=True):
        """Tag the *sentence* with the FEATS-3 tagger.

        :param sentence: sentence in Parsed CoNLL-U format; UPOS and LEMMA
                         fields must be already filled
        :type sentence: list(dict)
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param with_s_backoff: if result of separate FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_feats2(joint=False)``
        :type max_s_repeats: int
        :param with_j_backoff: if result of joint FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_feats2(joint=True)``
        :type max_j_repeats: int
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        sent = self.predict_feats2(sent, joint=False,
            with_backoff=with_s_backoff, max_repeats=max_s_repeats,
            inplace=True
        )
        sent_j = iter(self.predict_feats2(sent, joint=True,
            with_backoff=with_j_backoff, max_repeats=max_j_repeats,
            inplace=False
        ))
        for token in sent:
            feats = token['FEATS']
            feats_j = next(sent_j)['FEATS']
            for feat, val in feats_j.items():
                if feat not in feats:
                    feats[feat] = val
            del_feats = []
            for feat, val in feats.items():
                if feat not in feats_j:
                    del_feats.append(feat)
            for feat in del_feats:
                feats.pop(feat, None)
        return sentence

    def predict3(self, sentence, pos_backoff=True, pos_repeats=0,
                 feats_s_backoff=True, feats_s_repeats=0,
                 feats_j_backoff=True, feats_j_repeats=0, inplace=True):
        """Tag the *sentence* with the all available taggers.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: parameter for ``predict_pos2()``
        :type pos_repeats: int
        :type feats_s_backoff: if result of separate FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_s_repeats: parameter for ``predict_feats3()``
        :type feats_s_repeats: int
        :type feats_j_backoff: if result of joint FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_j_repeats: parameter for ``predict_feats3()``
        :type feats_j_repeats: int
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged *sentence* in Parsed CoNLL-U format
        """
        return \
            self.predict_feats3(
                self.predict_lemma(
                    self.predict_pos2(
                        sentence, with_backoff=pos_backoff,
                        max_repeats=pos_repeats, inplace=inplace
                    ),
                    inplace=inplace
                ),
                with_s_backoff=feats_s_backoff, max_s_repeats=feats_s_repeats,
                with_j_backoff=feats_j_backoff, max_j_repeats=feats_j_repeats,
                inplace=inplace
           )

    def predict_feats3_sents(self, sentences=None,
                             with_s_backoff=True, max_s_repeats=0,
                             with_j_backoff=True, max_j_repeats=0,
                             inplace=True, save_to=None):
        """Apply ``self.predict_feats2()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :type with_s_backoff: if result of separate FEATS-2 tagger differs
                              from both its base taggers, get one of the bases
                              on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_feats3()``
        :type max_s_repeats: int
        :type with_j_backoff: if result of joint FEATS-2 tagger differs
                              from both its base taggers, get one of the bases
                              on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_feats3()``
        :type max_j_repeats: int
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, the new list of
                        sentences will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_feats3(s,
                                     with_s_backoff=with_s_backoff,
                                     max_s_repeats=max_s_repeats,
                                     with_j_backoff=with_j_backoff,
                                     max_j_repeats=max_j_repeats,
                                     inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def predict3_sents(self, sentences=None,
                       pos_backoff=True, pos_repeats=0,
                       feats_s_backoff=True, feats_s_repeats=0,
                       feats_j_backoff=True, feats_j_repeats=0,
                       inplace=True, save_to=None):
        """Apply ``self.predict2()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: parameter for ``predict3()``
        :type feats_s_backoff: if result of separate FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_s_repeats: parameter for ``predict3()``
        :type feats_s_repeats: int
        :type feats_j_backoff: if result of joint FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_j_repeats: parameter for ``predict3()``
        :type feats_j_repeats: int
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict3(s,
                               pos_backoff=pos_backoff,
                               pos_repeats=pos_repeats,
                               feats_s_backoff=feats_s_backoff,
                               feats_s_repeats=feats_s_repeats,
                               feats_j_backoff=feats_j_backoff,
                               feats_j_repeats=feats_j_repeats,
                               inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def evaluate_feats3(self, gold=None, test=None,
                        with_s_backoff=True, max_s_repeats=0,
                        with_j_backoff=True, max_j_repeats=0,
                        feat=None, unknown_only=False, silent=False):
        """Score the accuracy of the FEATS-2 tagger against the *gold*
        standard. Remove feats (or only one specified feat) from the *gold*
        standard text, generate new feats using the tagger, then compute the
        accuracy score. If *test* is not None, compute the accuracy of the
        *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param with_s_backoff: if result of separate FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_feats3()``
        :type max_s_repeats: int
        :param with_j_backoff: if result of joint FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_feats3()``
        :type max_j_repeats: int
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :param silent: suppress log
        :return: accuracy scores of the tagger against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not;
                 2. by tags: sum of correctly detected feats to sum of all
                    feats that are non-empty in either gold or retagged 
                    sentence
        :rtype: tuple(float, float)
        """
        f = self.predict_feats
        self.predict_feats = \
            lambda sentence, rev=None, joint=None, feat=None, inplace=True: \
                self.predict_feats3(sentence,
                    with_s_backoff=with_s_backoff, max_s_repeats=max_s_repeats,
                    with_j_backoff=with_j_backoff, max_j_repeats=max_j_repeats,
                    inplace=inplace
                )
        res = self.evaluate_feats(gold=gold, test=test, feat=feat,
                                  unknown_only=unknown_only, silent=silent)
        self.predict_feats = f
        return res

    def evaluate3(self, gold=None, test=None,
                  pos_backoff=True, pos_repeats=0,
                  feats_s_backoff=True, feats_s_repeats=0,
                  feats_j_backoff=True, feats_j_repeats=0,
                  feat=None, unknown_only=False, silent=False):
        """Score a joint accuracy of the all available taggers against the
        *gold* standard. Extract wforms from the *gold* standard text, retag it
        using all the taggers, then compute a joint accuracy score.  If *test*
        is not None, compute the accuracy of the *test* corpus with respect to
        the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :type pos_backoff: if result of POS-2 tagger differs from both its
                           base taggers, get one of the bases on the ground
                           of some heuristics
        :param pos_repeats: parameter for ``predict3()``
        :type feats_s_backoff: if result of separate FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_s_repeats: parameter for ``predict3()``
        :type feats_s_repeats: int
        :type feats_j_backoff: if result of joint FEATS-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param feats_j_repeats: parameter for ``predict3()``
        :type feats_j_repeats: int
        :param feat: name of the feat to evaluate the tagger; if None, then
                     tagger will be evaluated for all feats
        :type feat: str
        :param silent: suppress log
        :param unknown_only: calculate accuracy score only for words that are
                             not present in train corpus
        :return: joint accuracy scores of the taggers against the gold:
                 1. by tokens: the tagging of the whole token may be either
                    correct or not
                 2. by tags: sum of correctly detected tags to sum of all tags
                    that are non-empty in either gold or retagged sentences
        :rtype: tuple(float, float)
        """
        f = self.predict
        self.predict = \
            lambda sentence, pos_rev=None, \
                   feats_joint=None, feats_rev=None, inplace=False: \
                self.predict3(sentence,
                              pos_backoff=pos_backoff,
                              pos_repeats=pos_repeats,
                              feats_s_backoff=feats_s_backoff,
                              feats_s_repeats=feats_s_repeats,
                              feats_j_backoff=feats_j_backoff,
                              feats_j_repeats=feats_j_repeats,
                              inplace=inplace)
        res = self.evaluate(gold=gold, test=test, feat=feat,
                            unknown_only=unknown_only, silent=silent)
        self.predict = f
        return res



# -*- coding: utf-8 -*-
# Morra project: Named Entity tagger
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Adjusting Morphological parser's models for Named Entity Recognition. It
includes all modification of the algorithm: forward and backward, joint and
separate, combining joint and separate, and, finally, the NER-3 algorithm
that make prediction on the ground of all its predecessors.
"""
from copy import deepcopy
import pickle
import random
import sys

from corpuscula.corpus_utils import _AbstractCorpus
from corpuscula.utils import LOG_FILE, print_progress, vote
from morra.base_parser import _AveragedPerceptron, BaseParser
from morra.features_ne import FeaturesNE


class MorphParserNE(BaseParser):
    """Named Entity parser"""

    def __init__(self, features='RU', guess_ne=None):
        super().__init__()
        self.features = FeaturesNE(lang=features) \
            if isinstance(features, str) else features

        self._guess_ne = guess_ne

        self._ne_freq       = None  # [(ne, cnt, freq)], sorted by freq
        self._ne_model      = None
        self._ne_rev_model  = None
        self._ne2_model     = None
        self._ne_models     = {}
        self._ne_rev_models = {}
        self._ne2_models    = {}

    def backup(self):
        """Get current state"""
        o = super().backup()
        o.update({'ne_freq'              : self._ne_freq,
                  'ne_model_weights'     : self._ne_model.weights
                                               if self._ne_model else
                                           None,
                  'ne_rev_model_weights' : self._ne_rev_model.weights
                                               if self._ne_rev_model else
                                           None,
                  'ne2_model_weights'    : self._ne2_model.weights
                                               if self._ne2_model else
                                           None,
                  'ne_models_weights'    : {
                      x: y.weights for x, y in self._ne_models.items()
                  },
                  'ne_rev_models_weights': {
                      x: y.weights for x, y in self._ne_rev_models.items()
                  },
                  'ne2_models_weights'   : {
                      x: y.weights for x, y in self._ne2_models.items()
                  }})
        return o

    def restore(self, o):
        """Restore current state from backup object"""
        super().restore(o)
        (self._ne_freq        ,
         ne_model_weights     ,
         ne_rev_model_weights ,
         ne2_model_weights    ,
         ne_models_weights    ,
         ne_rev_models_weights,
         ne2_models_weights   ) = [o.get(x) for x in ['ne_freq'              ,
                                                      'ne_model_weights'     ,
                                                      'ne_rev_model_weights' ,
                                                      'ne2_model_weights'    ,
                                                      'ne_models_weights'    ,
                                                      'ne_rev_models_weights',
                                                      'ne2_models_weights'   ]]
        if ne_model_weights:
            self._ne_model = _AveragedPerceptron()
            self._ne_model.weights = ne_model_weights
        else:
            self._ne_model = None
        if ne_rev_model_weights:
            self._ne_rev_model = _AveragedPerceptron()
            self._ne_rev_model.weights = ne_rev_model_weights
        else:
            self._ne_rev_model = None
        if ne2_model_weights:
            self._ne2_model = _AveragedPerceptron()
            self._ne2_model.weights = ne2_model_weights
        else:
            self._ne2_model = None
        models = self._ne_models = {}
        if ne_models_weights:
            for ne, weights in ne_models_weights.items():
                model = models[ne] = _AveragedPerceptron()
                model.weights = weights
        models = self._ne_rev_models = {}
        if ne_rev_models_weights:
            for ne, weights in ne_rev_models_weights.items():
                model = models[ne] = _AveragedPerceptron()
                model.weights = weights
        models = self._ne2_models = {}
        if ne2_models_weights:
            for ne, weights in ne2_models_weights.items():
                model = models[ne] = _AveragedPerceptron()
                model.weights = weights

    def _save_ne_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump((self._ne_freq,
                         self._ne_model.weights if self._ne_model else
                         None), f, 2)

    def _load_ne_model(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, weights = pickle.load(f)
            if weights:
                self._ne_model = _AveragedPerceptron()
                self._ne_model.weights = weights
            else:
                self._ne_model = None

    def _save_ne_rev_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump((self._ne_freq,
                         self._ne_rev_model.weights if self._ne_rev_model else
                         None), f, 2)

    def _load_ne_rev_model(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, weights = pickle.load(f)
            if weights:
                self._ne_rev_model = _AveragedPerceptron()
                self._ne_rev_model.weights = weights
            else:
                self._ne_rev_model = None

    def _save_ne2_model(self, file_path):
        with open(file_path, 'wb') as f:
            pickle.dump((self._ne_freq,
                         self._ne2_model.weights if self._ne2_model else
                         None), f, 2)

    def _load_ne2_model(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, weights = pickle.load(f)
            if weights:
                self._ne2_model = _AveragedPerceptron()
                self._ne2_model.weights = weights
            else:
                self._ne2_model = None

    def _save_ne_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump((
                self._ne_freq,
                (feat, self._ne_models[feat].weights) if feat else
                {x: y.weights for x, y in self._ne_models.items()}
            ), f, 2)

    def _load_ne_models(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._ne_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._ne_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def _save_ne_rev_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump((
                self._ne_freq,
                (feat, self._ne_rev_models[feat].weights) if feat else
                {x: y.weights for x, y in self._ne_rev_models.items()}
            ), f, 2)

    def _load_ne_rev_models(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._ne_rev_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._ne_rev_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def _save_ne2_models(self, file_path, feat=None):
        with open(file_path, 'wb') as f:
            pickle.dump((
                self._ne_freq,
                (feat, self._ne2_models[feat].weights) if feat else
                {x: y.weights for x, y in self._ne2_models.items()}
            ), f, 2)

    def _load_ne2_models(self, file_path):
        with open(file_path, 'rb') as f:
            self._ne_freq, o = pickle.load(f)
            if isinstance(o, tuple):
                feat, weights = o
                model = self._ne2_models[feat] = _AveragedPerceptron()
                model.weights = weights
            else:
                models = self._ne2_models = {}
                for feat, weights in o.items():
                    model = models[feat] = _AveragedPerceptron()
                    model.weights = weights

    def predict_ne(self, sentence, joint=True, rev=False, ne=None,
                   inplace=True, no_final_clean=False):
        """Tag the *sentence* with the NE tagger.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param joint: if True, use joint NE model (default); elsewise, use
                      separate models
        :param rev: if True, use Reversed NE tagger instead of generic
                    straight one
        :param ne: predict only specified NE. Can be used only with joint=False
        :type ne: str
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :param no_final_clean: don't search and remove NE from empty nodes
        :return: tagged sentence in Parsed CoNLL-U format
        """
        return (self._predict_ne_joint if joint else
                self._predict_ne_separate)(sentence, rev=rev, ne=ne,
                                           inplace=inplace)

    def predict_ne2(self, sentence, joint=True, with_backoff=True,
                    max_repeats=0, ne=None, inplace=True, no_final_clean=False):
        """Tag the *sentence* with the NE-2 tagger.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param joint: if True, use joint NE-2 model (default); elsewise, use
                      separate models
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one repeat
                            only for tokens where NE-1 taggers don't concur
        :type max_repeats: int
        :param ne: predict only specified NE. Can be used only with joint=False
        :type ne: str
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :param no_final_clean: don't search and remove NE from empty nodes
        :return: tagged sentence in Parsed CoNLL-U format
        """
        return (
            self._predict_ne2_joint if joint else
            self._predict_ne2_separate
        )(
            sentence, with_backoff=with_backoff, max_repeats=max_repeats,
            ne=ne, inplace=inplace
        )

    def _predict_ne_joint(self, sentence, rev=False,
                          ne=None, inplace=True, no_final_clean=False):
        assert not ne, 'ERROR: ne must be None with joint=True'
        cdict = self._cdict
        model = self._ne_rev_model if rev else self._ne_model
        assert model, 'ERROR: Use train_ne(joint=True{}) prior ' \
                      'to prepare {}NE tagger' \
                          .format(*((', rev=True', 'Reversed') if rev else
                                    ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        default_ne = '_'
        tokens = [(x['FORM'], x['LEMMA'],
                   x['UPOS'], x['FEATS'], x['MISC'])
                      for x in sent
                          if x['FORM'] and x['LEMMA'] and x['UPOS']
                         and '-' not in x['ID']]
        tokens_ = [[*x[:4], None] for x in tokens]
        if rev:
            tokens = tokens[::-1]
        context, lemma_context, pos_context, feats_context = \
            [list(x) for x in zip(*[t[:4] for t in tokens])] if tokens else \
            [[]] * 4
        prev, prev2 = self.features.START
        max_i = len(tokens) - 1
        for i, (wform, lemma, pos, feats, misc) in enumerate(tokens):
            i_ = max_i - i if rev else i
            guess, coef = None, None
            if self._guess_ne:
                guess, coef = self._guess_ne(None, None, i_,
                                             tokens_, cdict)
            if coef is not None and guess is None:
                guess = default_ne
            if coef != 1.:
                features = self.features.get_ne_features(
                    i, context, lemma_context, pos_context,
                    feats_context, prev, prev2
                )
                guess = model.predict(
                    features, suggest=guess, suggest_coef=coef
                )
            prev2 = prev
            tokens_[i_][4] = prev = guess
            if guess != default_ne:
                misc['NE'] = guess
            else:
                misc.pop('NE', None)
        if not no_final_clean:
            for misc in [
                x['MISC']
                    for x in sent
                        if not x['FORM'] or not x['LEMMA'] or not x['UPOS']
                       or '-' in x['ID']
            ]:
                misc.pop('NE', None)
        return sentence

    def _predict_ne_separate(self, sentence, rev=False,
                             ne=None, inplace=True, no_final_clean=None):
        cdict = self._cdict
        models = self._ne_rev_models if rev else self._ne_models
        assert models, 'ERROR: Use train_ne(joint=False{}) prior ' \
                       'to prepare {}NE tagger' \
                           .format(*((', rev=True', 'Reversed') if rev else
                                     ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        for token in sent:
            misc = token['MISC']
            if 'NE' in misc and (not ne or misc['NE'] == ne):
                del misc['NE']
        if not ne:
            for ne, _, _ in self._ne_freq:
                self._predict_ne_separate(sent, rev=rev, ne=ne, inplace=True)
        else:
            model = models[ne]
            default_ne = '_'
            tokens = [(x['FORM'], x['LEMMA'], x['UPOS'], x['FEATS'], x['MISC'])
                          for x in sent
                              if x['FORM'] and x['LEMMA'] and x['UPOS']
                             and '-' not in x['ID']]
            tokens_ = [[*x[:4], None] for x in tokens]
            if rev:
                tokens = tokens[::-1]
            context, lemma_context, pos_context, feats_context = \
                [list(x) for x in zip(*[t[:4] for t in tokens])] \
                    if tokens else \
                [[]] * 4
            prev, prev2 = self.features.START
            max_i = len(tokens) - 1
            for i, (wform, lemma, pos, feats, misc) in enumerate(tokens):
                i_ = max_i - i if rev else i
                guess, coef = None, None
                if self._guess_ne:
                    guess, coef = self._guess_ne(None, None, i_,
                                                 tokens_, cdict)
                if coef is not None:
                    guess = guess == ne
                if coef != 1.:
                    features = \
                        self.features.get_ne_features(
                            i, context, lemma_context, pos_context,
                            feats_context, prev, prev2
                        )
                    guess = model.predict(
                        features, suggest=guess, suggest_coef=coef
                    )
                prev2 = prev
                tokens_[i_][4] = prev = guess
                if guess:
                    misc['NE'] = ne
        return sentence

    def _predict_ne2_joint(self, sentence, with_backoff=True, max_repeats=0,
                           ne=None, inplace=True, no_final_clean=False):
        assert not ne, 'ERROR: ne must be None with joint=True'
        cdict = self._cdict
        model = self._ne2_model
        assert model, 'ERROR: Use train_ne2(joint=True) prior ' \
                      'to prepare NE-2 tagger'
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        default_ne = '_'
        sent = self._predict_ne_joint(sent, rev=True, ne=ne,
                                      inplace=True,
                                      no_final_clean=no_final_clean)
        sent_straight = self._predict_ne_joint(sent, rev=False, ne=ne,
                                               inplace=False,
                                               no_final_clean=True)
        tokens = [(x['FORM'], x['LEMMA'],
                   x['UPOS'], x['FEATS'], x['MISC'])
                      for x in sent
                          if x['FORM'] and x['LEMMA'] and x['UPOS']
                         and '-' not in x['ID']]
        tokens_straight = [(x['FORM'], x['LEMMA'],
                            x['UPOS'], x['FEATS'], x['MISC'])
                               for x in sent_straight
                                   if x['FORM'] and x['LEMMA'] and x['UPOS']
                                  and '-' not in x['ID']]
        context, lemma_context, pos_context, feats_context, ne_context_rev = \
            [list(x) for x in zip(*[[*t[:4], t[4].get('NE', default_ne)]
                                        for t in tokens])] if tokens else \
            [[]] * 5

        ne_context_straight = [t[4].get('NE', default_ne)
                                   for t in tokens_straight]
        # Rev model is better for initial word (with capital letter?)
        tokens_ = [[*t[:4], None] for t in tokens][:1] \
                + [[*t[:4], None] for t in tokens_straight][1:]
        ne_context = ne_context_rev[:1] + ne_context_straight[1:]
        ###
        changes = len(tokens) + 1
        i_ = 1
        while True:
            changes_prev = changes
            changes = 0
            ne_context_rev_i      = iter(ne_context_rev     )
            ne_context_straight_i = iter(ne_context_straight)
            for i, (wform, lemma, pos, feats, misc) in enumerate(tokens):
                ne_rev      = next(ne_context_rev_i     )
                ne_straight = next(ne_context_straight_i)
                if ne_rev != ne_straight or (not with_backoff
                                         and max_repeats > 0):
                    guess, coef = self._guess_ne(
                        None, None, i, tokens_, cdict
                    ) if self._guess_ne else (None, None)
                    if coef is not None and guess is None:
                        guess = default_ne
                    if coef != 1.:
                        features = self.features.get_ne2_features(
                            i, context, lemma_context, pos_context,
                            feats_context, ne_context
                        )
                        guess = model.predict(
                            features, suggest=guess, suggest_coef=coef
                        )
                        if with_backoff and guess not in [ne_rev, ne_straight]:
                            guess = ne_context[i]
                    if guess != misc.get('NE', default_ne):
                        changes += 1
                    tokens_[i][4] = ne_context[i] = guess
                    if guess != default_ne:
                        misc['NE'] = guess
                    else:
                        misc.pop('NE', None)
            if with_backoff or changes == 0:
                break
            elif changes > changes_prev:
                for token, token_prev in zip(tokens, tokens_prev):
                    tokens[4] = token_prev[4].copy()
                break
            if i_ >= max_repeats:
                break
            tokens_prev = deepcopy(tokens)
            i_ += 1
        return sentence

    def _predict_ne2_separate(self, sentence, with_backoff=True,
                              max_repeats=0, ne=None, inplace=True,
                              no_final_clean=None):
        cdict = self._cdict
        models = self._ne2_models
        assert models, \
               'ERROR: Use train_ne2(joint=False) prior ' \
               'to prepare NE-2 tagger' \
                   .format(*(('rev=True', 'Reversed') if rev else ('', '')))
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        if not ne:
            for ne, _, _ in self._ne_freq:
                self._predict_ne2_separate(sent, with_backoff=with_backoff,
                                           max_repeats=max_repeats,
                                           ne=ne, inplace=True)
        else:
            model = models[ne]
            default_ne = '_'
            sent = self._predict_ne_separate(sent, rev=True, ne=ne,
                                             inplace=True)
            sent_straight = self._predict_ne_separate(sent, rev=False, ne=ne,
                                                      inplace=False)
            tokens = [(x['FORM'], x['LEMMA'],
                       x['UPOS'], x['FEATS'], x['MISC'])
                          for x in sent
                              if x['FORM'] and x['LEMMA'] and x['UPOS']
                             and '-' not in x['ID']]
            tokens_straight = [(x['FORM'], x['LEMMA'],
                                x['UPOS'], x['FEATS'], x['MISC'])
                                   for x in sent_straight
                                       if x['FORM'] and x['LEMMA']
                                      and x['UPOS'] and '-' not in x['ID']]
            context, lemma_context, pos_context, feats_context, \
                                                            ne_context_rev = \
                [list(x) for x in zip(*[[*t[:4], t[4].get('NE', default_ne)]
                                            for t in tokens])] if tokens else \
                [[]] * 5
            ne_context_straight = [t[4].get('NE', default_ne)
                                       for t in tokens_straight]
            # Rev model is better for initial word (with capital letter?)
            tokens_ = [[*t[:4], None] for t in tokens][:1] \
                    + [[*t[:4], None] for t in tokens_straight][1:]
            ne_context = ne_context_rev[:1] + ne_context_straight[1:]
            ###
            changes = len(tokens) + 1
            i_ = 1
            while True:
                changes_prev = changes
                changes = 0
                ne_context_rev_i      = iter(ne_context_rev     )
                ne_context_straight_i = iter(ne_context_straight)
                for i, (wform, lemma, pos, feats, misc) in enumerate(tokens):
                    ne_rev      = next(ne_context_rev_i     )
                    ne_straight = next(ne_context_straight_i)
                    if ne_rev != ne_straight or (not with_backoff
                                             and max_repeats > 0):
                        guess, coef = None, None
                        if self._guess_ne:
                            guess, coef = self._guess_ne(None, None, i,
                                                         tokens_, cdict)
                        if coef is not None:
                            guess = guess == ne
                        if coef != 1.:
                            features = self.features.get_ne2_features(
                                i, context, lemma_context, pos_context,
                                feats_context, ne_context
                            )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef
                            )
                            if with_backoff and guess not in [ne_rev,
                                                              ne_straight]:
                                guess = ne_context[i]
                        if guess != misc.get('NE', default_ne):
                            changes += 1
                        tokens_[i][4] = ne_context[i] = guess
                        if guess:
                            misc['NE'] = ne_context[i] = ne
                if with_backoff or changes == 0:
                    break
                elif changes > changes_prev:
                    for token, token_prev in zip(tokens, tokens_prev):
                        tokens[4] = token_prev[4].copy()
                    break
                if i_ >= max_repeats:
                    break
                tokens_prev = deepcopy(tokens)
                i_ += 1
        return sentence

    def predict_ne3(self, sentence,
                    with_s_backoff=True, max_s_repeats=0,
                    with_j_backoff=True, max_j_repeats=0,
                    inplace=True):
        """Tag the *sentence* with the NE-3 tagger.

        :param sentence: sentence in Parsed CoNLL-U format
        :type sentence: list(dict)
        :param with_s_backoff: if result of the separate NE-1 tagger differs
                               from both base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_ne2(joint=False)``
        :type max_s_repeats: int
        :param with_j_backoff: if result of the joint NE-1 tagger differs
                               from both base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_ne2(joint=True)``
        :type max_j_repeats: int
        :param inplace: if True, method changes and returns the given sentence
                        itself; elsewise, new sentence will be created
        :return: tagged sentence in Parsed CoNLL-U format
        """
        if not inplace:
            sentence = deepcopy(sentence)
        sent = sentence[0] if isinstance(sentence, tuple) else sentence
        sent = self._predict_ne2_separate(
            sent, with_backoff=with_s_backoff, max_repeats=max_s_repeats,
            inplace=True
        )
        sent_j = iter(self._predict_ne2_joint(
            sent, with_backoff=with_j_backoff, max_repeats=max_j_repeats,
            inplace=False, no_final_clean=True))
        for i, token in enumerate(sent):
            ne = token['MISC'].get('NE')
            ne_j = next(sent_j)['MISC'].get('NE')
            if not ne:
                if ne_j:
                    token['MISC']['NE'] = ne_j
            elif not ne_j:
                token['MISC'].pop('NE', None)
        return sentence

    def predict_ne_sents(self, sentences=None, joint=True, rev=False,
                         ne=None, inplace=True, save_to=None):
        """Apply ``self.predict_ne()`` to each element of *sentences*.

        :param sentences: a name of a file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param joint: if True, use joint NE model (default); elsewise, use
                      separate models
        :param rev: if True, use Reversed NE tagger instead of generic
                    straight one
        :param ne: name of the NE to tag; if None, then all possible NEs will
                   be tagged. Must be None if joint=True
        :type ne: str
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_ne(
                     s, joint=joint, rev=rev, ne=ne, inplace=inplace
                 )
                     for s in sentences),
            save_to=save_to
        )

    def predict_ne2_sents(self, sentences=None, joint=True, with_backoff=True,
                          max_repeats=0, ne=None, inplace=True, save_to=None):
        """Apply ``self.predict_ne2()`` to each element of *sentences*.

        :param sentences: a name of a file in CoNLL-U format or list/iterator
                          of sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param joint: if True, use joint NE-2 model (default); elsewise, use
                      separate models
        :param with_backoff: if result of the tagger differs from both base
                             taggers, get one of the bases on the ground of
                             some heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one repeat
                            only for tokens where NE-1 taggers don't concur
        :type max_repeats: int
        :param ne: name of the NE to tag; if None, then all possible NEs will
                   be tagged. Must be None if joint=True
        :type ne: str
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_ne2(
                     s, joint=joint, with_backoff=with_backoff,
                     max_repeats=max_repeats, ne=ne, inplace=inplace
                 )
                     for s in sentences),
            save_to=save_to
        )

    def predict_ne3_sents(self, sentences=None, inplace=True,
                          with_s_backoff=True, max_s_repeats=0,
                          with_j_backoff=True, max_j_repeats=0,
                          save_to=None):
        """Apply ``self.predict_ne3()`` to each element of *sentences*.

        :param sentences: a name of file in CoNLL-U format or list/iterator of
                          sentences in Parsed CoNLL-U. If None, then loaded
                          test corpus is used
        :param with_s_backoff: if result of the separate NE-1 tagger differs
                               from both base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_ne3()``
        :type max_s_repeats: int
        :param with_j_backoff: if result of the joint NE-1 tagger differs
                               from both base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_ne3()``
        :type max_j_repeats: int
        :param inplace: if True, method changes and returns the given
                        sentences themselves; elsewise, new list of sentences
                        will be created
        :param save_to: if not None then the result will be saved to the file
                        with a specified name
        :type save_to: str
        """
        return self._predict_sents(
            sentences,
            lambda sentences:
                (self.predict_ne3(s,
                                  with_s_backoff=with_s_backoff,
                                  max_s_repeats=max_s_repeats,
                                  with_j_backoff=with_j_backoff,
                                  max_j_repeats=max_j_repeats,
                                  inplace=inplace)
                     for s in sentences),
            save_to=save_to
        )

    def evaluate_ne(self, gold=None, test=None, joint=True, rev=False,
                    ne=None, silent=False):
        """Score the accuracy of the NE tagger against the *gold* standard.
        Remove NE tags from the *gold* standard text, retag it using the
        tagger, then compute the accuracy score. If *test* is not None, compute
        the accuracy of the *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param joint: if True, use joint NE model (default); elsewise, use
                      separate models
        :param rev: if True, use Reversed NE tagger instead of generic
                    straight one
        :param ne: name of the NE to evaluate; if None, then all possible NEs
                   will be evaluated
        :type ne: str
        :param silent: suppress log
        :return: accuracy score of the tagger against the gold
        :rtype: float
        """
        nn = cc = n = c = w = w1 = w2 = 0
        if gold is None:
            gold = self._test_corpus
        elif (isinstance(gold, type) and issubclass(gold, _AbstractCorpus)) \
          or isinstance(gold, _AbstractCorpus):
            gold = gold.test()
        assert gold, 'ERROR: Gold must not be empty'
        corpus_len = len(gold) if isinstance(gold, list) else None
        gold = self._get_corpus(gold)
        if test:
            test = self._get_corpus(test, silent=True)
        header = '{}NE{}'.format('Reversed ' if rev else '',
                                 '<<{}>>'.format(ne) if ne else '')
        if not silent:
            print('Evaluate ' + header, file=LOG_FILE)
        progress_step = max(int(corpus_len / 60), 1000) \
                            if corpus_len else 1000
        progress_check_step = min(int(corpus_len / 100), 1000) \
                                  if corpus_len else 100
        i = -1
        for i, gold_sent in enumerate(gold):
            if not silent and not i % progress_check_step:
                print_progress(i, end_value=corpus_len, step=progress_step,
                               file=LOG_FILE)
            test_sent = next(test) if test else \
                        self.predict_ne(gold_sent, joint=joint, rev=rev,
                                        ne=None if joint else ne,
                                        inplace=False)
            for j, gold_token in enumerate(gold_sent):
                if gold_token['FORM'] and '-' not in gold_token['ID']:
                    gold_ne = gold_token['MISC'].get('NE')
                    test_ne = test_sent[j]['MISC'].get('NE')
                    nn += 1
                    if (ne and (gold_ne == ne or test_ne == ne)) \
                    or (not ne and (gold_ne or test_ne)):
                        n += 1
                        if gold_ne == test_ne:
                            c += 1
                            cc += 1
                        elif not gold_ne or (ne and gold_ne != ne):
                            w1 += 1
                        elif not test_ne \
                          or (ne and test_ne != ne):
                            w2 += 1
                        else:
                            w += 1
                    else:
                        cc += 1
        if not silent:
            if i < 0:
                print('Nothing to do!', file=LOG_FILE)
            else:
                print_progress(i + 1,
                               end_value=corpus_len if corpus_len else 0,
                               step=progress_step, file=LOG_FILE)
                sp = ' ' * (len(header) - 2)
                print(header + ' total: {}'.format(n), file=LOG_FILE)
                print(sp   + ' correct: {}'.format(c), file=LOG_FILE)
                print(sp   + '   wrong: {} [{} excess / {} absent{}]'
                                  .format(n - c, w1, w2, '' if ne else
                                                         ' / {} wrong type'
                                                             .format(w)),
                      file=LOG_FILE)
                print(sp   + 'Accuracy: {}'.format(c / n if n > 0 else 1.))
                print('[Total accuracy: {}]'.format(cc / nn if nn > 0 else 1.))
        return c / n if n > 0 else 1.

    def evaluate_ne2(self, gold=None, test=None, joint=True,
                     with_backoff=True, max_repeats=0, ne=None, silent=False):
        """Score the accuracy of the NE-2 tagger against the *gold* standard.
        Remove NE tags from the *gold* standard text, retag it using the
        tagger, then compute the accuracy score. If *test* is not None, compute
        the accuracy of the *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param joint: if True, use joint NE-2 model (default); elsewise, use
                      separate models
        :type with_backoff: if result of the tagger differs from both base
                            taggers, get one of the bases on the ground of some
                            heuristics
        :param max_repeats: repeat a prediction step based on the previous one
                            while changes in prediction are diminishing and
                            ``max_repeats`` is not reached. 0 means one repeat
                            only for tokens where NE-1 taggers don't concur
        :type max_repeats: int
        :param ne: name of the NE to evaluate; if None, then all possible NEs
                   will be evaluated
        :type ne: str
        :param silent: suppress log
        :return: accuracy score of the tagger against the gold
        :rtype: float
        """
        f = self.predict_ne
        self.predict_ne = \
            lambda sentence, joint=joint, rev=None, ne=ne, inplace=True, \
                   no_final_clean=False: \
                self.predict_ne2(sentence, joint=joint,
                                 with_backoff=with_backoff,
                                 max_repeats=max_repeats, ne=ne,
                                 inplace=inplace,
                                 no_final_clean=no_final_clean)
        res = self.evaluate_ne(gold=gold, test=test, joint=joint, ne=ne,
                               silent=silent)
        self.predict_ne_sents = f
        return res

    def evaluate_ne3(self, gold=None, test=None,
                     with_s_backoff=True, max_s_repeats=0,
                     with_j_backoff=True, max_j_repeats=0, silent=False):
        """Score the accuracy of the NE-3 tagger against the *gold* standard.
        Remove NE tags from the *gold* standard text, retag it using the
        tagger, then compute the accuracy score. If *test* is not None, compute
        the accuracy of the *test* corpus with respect to the *gold*.

        :param gold: a corpus of tagged sentences to score the tagger on.
                     If *gold* is None then loaded test corpus is used
        :param test: a corpus of tagged sentences to compare with *gold*
        :param with_s_backoff: if result of separate NE-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_s_repeats: parameter for ``predict_ne3()``
        :type max_s_repeats: int
        :param with_j_backoff: if result of joint NE-2 tagger differs
                               from both its base taggers, get one of the bases
                               on the ground of some heuristics
        :param max_j_repeats: parameter for ``predict_ne3()``
        :type max_j_repeats: int
        :param silent: suppress log
        :return: accuracy score of the tagger against the gold
        :rtype: float
        """
        f = self.predict_ne
        self.predict_ne = \
            lambda sentence, joint=None, rev=None, ne=None, inplace=True, \
                   no_final_clean=None: \
                self.predict_ne3(sentence,
                                 with_s_backoff=with_s_backoff,
                                 max_s_repeats=max_s_repeats,
                                 with_j_backoff=with_j_backoff,
                                 max_j_repeats=max_j_repeats,
                                 inplace=inplace)
        res = self.evaluate_ne(gold=gold, test=test, silent=silent)
        self.predict_ne_sents = f
        return res

    def train_ne(self, joint=True, rev=False, ne=None, epochs=5,
                 no_train_evals=True, seed=None, dropout=None,
                 context_dropout=None):
        """Train a NE tagger from ``self._train_corpus``.

        :param joint: if True, train joint NE model (default); elsewise, train
                      separate models
        :param rev: if True, train Reversed NE tagger instead of generic
                    straight one
        :param ne: name of the entity to evaluate the tagger; if None
                   (default), then tagger will be evaluated for all entities.
        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of NE tags to be randomly replaced
                                after predict to random NE tags to prevent
                                overfitting
        """
        return (self._train_ne_joint if joint else
                self._train_ne_separate)(rev=rev, ne=ne, epochs=epochs,
                                         no_train_evals=no_train_evals,
                                         seed=seed)

    def train_ne2(self, joint=True, ne=None, epochs=5,
                  test_max_repeats=0, no_train_evals=True, seed=None,
                  dropout=None, context_dropout=None):
        """Train a NE-2 tagger from ``self._train_corpus``.

        :param joint: if True, train joint NE-2 model (default); elsewise,
                      train separate models
        :param epochs: number of training iterations. If epochs < 0, then the
                       best model will be searched based on evaluation of test
                       corpus. The search will be stopped when the result of
                       next |epochs| iterations will be worse than the best
                       one. It's allowed to specify epochs as tuple of both
                       variants (positive and negative)
        :type epochs: int|tuple(int, int)
        :param test_max_repeats: parameter for ``evaluate_ne2()``
        :type test_max_repeats: int
        :param no_train_evals: don't make interim and final evaluations on the
                               training set (save time)
        :param seed: init value for the random number generator
        :type seed: int
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict to prevent overfitting
        :type dropout: float
        :param context_dropout: a fraction of NE tags to be randomly replaced
                                after predict to random NE tags to prevent
                                overfitting
        """
        return (
            self._train_ne2_joint if joint else
            self._train_ne2_separate
        )(
            ne=ne, epochs=epochs,
            no_train_evals=no_train_evals, test_max_repeats=test_max_repeats,
            seed=seed
        )

    def _train_ne_joint(self, rev=False, ne=None, epochs=5,
                        no_train_evals=True, seed=None,
                        dropout=None, context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                                                           epochs, epochs_ = \
                        self._train_init(epochs, seed, allow_empty_cdict=True)
        assert not ne, 'ERROR: ne must be None with joint=True'

        default_ne = '_'

        model = _AveragedPerceptron(default_class=default_ne)
        if rev:
            self._ne_rev_model = model
        else:
            self._ne_model = model

        self._ne_freq = vote(x['MISC'].get('NE')
                                 for x in self._train_corpus for x in x
                                     if x['FORM'] and x['LEMMA'] and x['UPOS']
                                    and '-' not in x['ID']
                                    and x['MISC'].get('NE'))
        ne_classes = sorted(x[0] for x in self._ne_freq)

        header = '{}NE'.format('Reversed ' if rev else '')
        last_class_idx = len(ne_classes)
        print(ne_classes, file=LOG_FILE)
        ne_classes.append(default_ne)
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['LEMMA'],
                           x['UPOS'], x['FEATS'], x['MISC'])
                              for x in sentence
                                  if x['FORM'] and x['LEMMA'] and x['UPOS']
                                 and '-' not in x['ID']]
                tokens_ = [[*x[:4], None] for x in tokens]
                if rev:
                    tokens = tokens[::-1]
                context, lemma_context, pos_context, feats_context = \
                    [list(x) for x in zip(*[t[:4] for t in tokens])] \
                        if tokens else \
                    [[]] * 4
                prev, prev2 = self.features.START

                max_i = len(tokens) - 1
                for i, (wform, pos, lemma, feats, misc) in enumerate(tokens):
                    i_ = max_i - i if rev else i
                    gold_ne = misc.get('NE', default_ne)
                    guess, coef = None, None
                    if self._guess_ne:
                        guess, coef = self._guess_ne(None, None, i_,
                                                     tokens_, cdict)
                    if coef is not None:
                        if guess is None:
                            guess = default_ne
                        if guess == gold_ne:
                            td2 += 1
                        else:
                            fd2 += 1
                    if coef == 1.:
                        if guess == gold_ne:
                            td += 1
                        else:
                            fd += 1
                    else:
                        features = self.features.get_ne_features(
                            i, context, lemma_context, pos_context,
                            feats_context, prev, prev2
                        )
                        guess = model.predict(
                            features, suggest=guess, suggest_coef=coef,
                            dropout=dropout
                        )
                        if guess == gold_ne:
                            tp += 1
                        else:
                            fp += 1
                        model.update(gold_ne, guess, features)
                    if guess != default_ne or gold_ne != default_ne:
                        n += 1
                        c += guess == gold_ne

                    prev2 = prev
                    tokens_[i_][4] = prev = \
                        guess if not context_dropout \
                              or rand() >= context_dropout else \
                        ne_classes[randint(0, last_class_idx)]

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    self.evaluate_ne, {'joint': True, 'rev': rev}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals,
            self.evaluate_ne, {'joint': True, 'rev': rev}
        )

    def _train_ne_separate(self, rev=False, ne=None, epochs=5,
                           no_train_evals=True, seed=None, dropout=None,
                           context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                                                           epochs, epochs_ = \
                        self._train_init(epochs, seed, allow_empty_cdict=True)

        models = self._ne_rev_models if rev else self._ne_models

        self._ne_freq = vote(x['MISC'].get('NE')
                                 for x in self._train_corpus for x in x
                                     if x['FORM'] and x['LEMMA'] and x['UPOS']
                                    and '-' not in x['ID']
                                    and x['MISC'].get('NE'))
        ne_classes = [ne] if ne else [x[0] for x in self._ne_freq]

        for ne in [ne] if ne else sorted(ne_classes):
            model = models[ne] = _AveragedPerceptron()
            header = '{}NE<<{}>>'.format('Reversed ' if rev else '', ne)
            best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
            epoch = 0
            while True:
                n = c = 0
                td = fd = td2 = fd2 = tp = fp = 0
                random.shuffle(self._train_corpus)
                print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
                for sent_no, sentence in enumerate(self._train_corpus):
                    if not sent_no % progress_check_step:
                        print_progress(sent_no, end_value=corpus_len,
                                       step=progress_step)

                    tokens = [(x['FORM'], x['LEMMA'],
                               x['UPOS'], x['FEATS'], x['MISC'])
                                  for x in sentence
                                      if x['FORM'] and x['LEMMA'] and x['UPOS']
                                     and '-' not in x['ID']]
                    tokens_ = [[*t[:4], None] for t in tokens]
                    if rev:
                        tokens = tokens[::-1]
                    context, lemma_context, pos_context, feats_context = \
                        [list(x) for x in zip(*[t[:4] for t in tokens])] \
                            if tokens else \
                        [[]] * 4
                    prev, prev2 = self.features.START

                    max_i = len(tokens) - 1
                    for i, (wform, pos, lemma, feats, misc) in enumerate(tokens):
                        i_ = max_i - i if rev else i
                        gold_ne = misc.get('NE') == ne
                        guess, coef = None, None
                        if self._guess_ne:
                            guess, coef = self._guess_ne(None, None, i_,
                                                         tokens_, cdict)
                        if coef is not None:
                            guess = guess == ne
                            if guess == gold_ne:
                                td2 += 1
                            else:
                                fd2 += 1
                        if coef == 1.:
                            if guess == gold_ne:
                                td += 1
                            else:
                                fd += 1
                        else:
                            features = self.features.get_ne_features(
                                i, context, lemma_context, pos_context,
                                feats_context, prev, prev2
                            )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef,
                                dropout=dropout
                            )
                            if guess == gold_ne:
                                tp += 1
                            else:
                                fp += 1
                            model.update(gold_ne, guess, features)
                        if guess or gold_ne:
                            n += 1
                            c += guess == gold_ne

                        prev2 = prev
                        tokens_[i_][4] = prev = \
                            guess if not context_dropout \
                                  or rand() >= context_dropout else \
                            bool(randint(0, 1))

                print_progress(sent_no + 1, end_value=corpus_len,
                               step=progress_step)
                epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                    self._train_eval(
                        model, epoch, epochs, epochs_,
                        best_epoch, best_score, best_weights,
                        eqs, bads, score,
                        td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                        self.evaluate_ne,
                        {'joint': False, 'rev': rev, 'ne': ne}
                    )
                if eqs == -1:
                    break

            res = self._train_done(
                header, model, eqs, no_train_evals, self.evaluate_ne,
                {'joint': False, 'rev': rev, 'ne': ne}
            )

        return res if ne else \
               f_evaluate(joint=False, rev=rev, ne=ne, silent=True)

    def _train_ne2_joint(self, ne=None, epochs=5,
                         no_train_evals=True, test_max_repeats=0, seed=None,
                         dropout=None, context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                                                           epochs, epochs_ = \
                        self._train_init(epochs, seed, allow_empty_cdict=True)
        assert not ne, 'ERROR: ne must be None with joint=True'

        default_ne = '_'

        model = self._ne2_model = _AveragedPerceptron(default_class=default_ne)

        ne_classes = sorted(x[0] for x in self._ne_freq)

        header = 'NE-2'
        last_class_idx = len(ne_classes)
        print(ne_classes, file=LOG_FILE)
        ne_classes.append(default_ne)
        best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
        epoch = 0
        while True:
            n = c = 0
            td = fd = td2 = fd2 = tp = fp = 0
            random.shuffle(self._train_corpus)
            print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
            for sent_no, sentence in enumerate(self._train_corpus):
                if not sent_no % progress_check_step:
                    print_progress(sent_no, end_value=corpus_len,
                                   step=progress_step)

                tokens = [(x['FORM'], x['LEMMA'],
                           x['UPOS'], x['FEATS'], x['MISC'])
                              for x in sentence
                                  if x['FORM'] and x['LEMMA'] and x['UPOS']
                                 and '-' not in x['ID']]
                context, lemma_context, pos_context, feats_context, \
                                                                ne_context = \
                    [list(x) for x in zip(*[[*t[:4], t[4].get('NE',
                                                              default_ne)]
                                                for t in tokens])] \
                        if tokens else \
                    [[]] * 5
                tokens_ = [[*t[:4], None] for t in tokens]

                for i, (wform, pos, lemma, feats, misc) in enumerate(tokens):
                    gold_ne = misc.get('NE', default_ne)
                    guess, coef = None, None
                    if self._guess_ne:
                        guess, coef = self._guess_ne(None, None, i,
                                                     tokens_, cdict)
                    if coef is not None:
                        if guess is None:
                            guess = default_ne
                        if guess == gold_ne:
                            td2 += 1
                        else:
                            fd2 += 1
                    if coef == 1.:
                        if guess == gold_ne:
                            td += 1
                        else:
                            fd += 1
                    else:
                        features = self.features.get_ne2_features(
                            i, context, lemma_context, pos_context,
                            feats_context, ne_context
                        )
                        guess = model.predict(
                            features, suggest=guess, suggest_coef=coef,
                            dropout=dropout
                        )
                        if guess == gold_ne:
                            tp += 1
                        else:
                            fp += 1
                        model.update(gold_ne, guess, features)
                    if guess != default_ne or gold_ne != default_ne:
                        n += 1
                        c += guess == gold_ne

                    tokens_[i][4] = ne_context[i] = \
                        guess if not context_dropout \
                              or rand() >= context_dropout else \
                        ne_classes[randint(0, last_class_idx)]

            print_progress(sent_no + 1, end_value=corpus_len,
                           step=progress_step)
            epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                self._train_eval(
                    model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    self.evaluate_ne2, {'joint': True, 'with_backoff': False,
                                        'max_repeats': test_max_repeats}
                )
            if eqs == -1:
                break

        return self._train_done(
            header, model, eqs, no_train_evals,
            self.evaluate_ne2, {'joint': True, 'with_backoff': False,
                                'max_repeats': test_max_repeats}
        )

    def _train_ne2_separate(self, ne=None, epochs=5,
                            no_train_evals=True, test_max_repeats=0,
                            seed=None, dropout=None, context_dropout=None):
        cdict, corpus_len, progress_step, progress_check_step, \
                                                           epochs, epochs_ = \
                        self._train_init(epochs, seed, allow_empty_cdict=True)

        default_ne = '_'

        models = self._ne2_models

        ne_classes = [ne] if ne else [x[0] for x in self._ne_freq]

        for ne in [ne] if ne else sorted(ne_classes):
            model = models[ne] = _AveragedPerceptron()
            header = 'NE-2<<{}>>'.format(ne)
            best_epoch, best_score, best_weights, eqs, bads, score = \
                                                        -1, -1, None, 0, 0, -1
            epoch = 0
            while True:
                n = c = 0
                td = fd = td2 = fd2 = tp = fp = 0
                random.shuffle(self._train_corpus)
                print('{} Epoch {}'.format(header, epoch), file=LOG_FILE)
                for sent_no, sentence in enumerate(self._train_corpus):
                    if not sent_no % progress_check_step:
                        print_progress(sent_no, end_value=corpus_len,
                                       step=progress_step)

                    tokens = [(x['FORM'], x['LEMMA'],
                               x['UPOS'], x['FEATS'], x['MISC'])
                                  for x in sentence
                                      if x['FORM'] and x['LEMMA'] and x['UPOS']
                                     and '-' not in x['ID']]
                    context, lemma_context, pos_context, feats_context, \
                                                                ne_context = \
                        [list(x) for x in zip(*[[*t[:4], t[4].get('NE',
                                                                  default_ne)]
                                                    for t in tokens])] \
                            if tokens else \
                        [[]] * 5
                    tokens_ = [[*t[:4], None] for t in tokens]

                    for i, (wform, pos, lemma, feats, misc) in enumerate(tokens):
                        gold_ne = misc.get('NE') == ne
                        guess, coef = None, None
                        if self._guess_ne:
                            guess, coef = self._guess_ne(None, None, i,
                                                         tokens_, cdict)
                        if coef is not None:
                            guess = guess == ne
                            if guess == gold_ne:
                                td2 += 1
                            else:
                                fd2 += 1
                        if coef == 1.:
                            if guess == gold_ne:
                                td += 1
                            else:
                                fd += 1
                        else:
                            features = self.features.get_ne2_features(
                                i, context, lemma_context, pos_context,
                                feats_context, ne_context
                            )
                            guess = model.predict(
                                features, suggest=guess, suggest_coef=coef,
                                dropout=dropout
                            )
                            if guess == gold_ne:
                                tp += 1
                            else:
                                fp += 1
                            model.update(gold_ne, guess, features)
                        if guess or gold_ne:
                            n += 1
                            c += guess == gold_ne

                        tokens_[i][4] = ne_context[i] = \
                            guess if not context_dropout \
                                  or rand() >= context_dropout else \
                            bool(randint(0, 1))

                print_progress(sent_no + 1, end_value=corpus_len,
                               step=progress_step)
                epoch, epochs, best_epoch, best_score, best_weights, \
                                                          eqs, bads, score = \
                    self._train_eval(
                        model, epoch, epochs, epochs_,
                        best_epoch, best_score, best_weights,
                        eqs, bads, score,
                        td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                        self.evaluate_ne2,
                        {'joint': False, 'with_backoff': False,
                         'max_repeats': test_max_repeats, 'ne': ne}
                    )
                if eqs == -1:
                    break

            res = self._train_done(
                header, model, eqs, no_train_evals, self.evaluate_ne2,
                {'joint': False, 'with_backoff': False,
                 'max_repeats': test_max_repeats, 'ne': ne}
            )

        return res if ne else \
               f_evaluate(joint=False, ne=ne, silent=True)



# -*- coding: utf-8 -*-
# Morra project: Features for MorphParser2
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
If you need MorphParser2 to support your language, add feature-functions for
your language here. Then, create parser as:

    ``MorphParser2(features='<your lang>')``
"""
from morra.features import Features


class Features2(Features):
    """Features for MorphParser2"""

    def __init__(self, lang='RU'):
        super().__init__(lang=lang)
        if lang == 'RU':
            self.get_pos2_features = self.get_pos2_features_RU
            self.get_feat2_features = self.get_feat2_features_RU

    def get_pos2_features_RU(self, i, context, pos_context):
        context     = self.START + context     + self.END
        pos_context = self.START + pos_context + self.END
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        #pos_i  = pos_context[i]
        pos_b1 = pos_context[i - 1]
        pos_b2 = pos_context[i - 2]
        #pos_b3 = pos_context[i - 3] if i >= 3 else '-START3-'
        pos_f1 = pos_context[i + 1]
        pos_f2 = pos_context[i + 2]
        #pos_f3 = pos_context[i + 3] if i < len(pos_context) - 3 else '-END3-'

        features = self.init_features()

        self.add_feature(features, '_')
        self.add_feature(features, 'w', wform)
        self.add_feature(features, 's', self.wform_shape(wform))
        len_ = len(wform_i)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform_i[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform_i[-val:])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform_i[len_ - start - val:len_ - start]) 
        for val in range(2, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform_i[val:])
        #for val in range(1, 8):
        #    self.add_feature(features, 'P' + str(val),
        #                     '' if val > len_ else wform_i[:-val])

        #for val in range(3, 7):
        #    for start in range(1, len_ - 1 - val):
        #        self.add_feature(features, 'i' + str(val) + ':' + str(start),
        #                         wform_i[start:start + val])
        #for val in range(2, 8):
        #    self.add_feature(features, 'I' + str(val), wform[val:-val])
        ##self.add_feature(features, '-3p', pos_b3)
        ##self.add_feature(features, '+3p', pos_f3)

        self.add_feature(features, '-1p', pos_b1)
        self.add_feature(features, '-2p', pos_b2)
        self.add_feature(features, '-1p-2p', pos_b1, pos_b2)
        #self.add_feature(features, '-1p-2p-3p', pos_b1, pos_b2, pos_b3)
        self.add_feature(features, '-1pw', pos_b1, wform_i)
        #self.add_feature(features, '-1p-2pw', pos_b1, pos_b2, wform_i)
        #self.add_feature(features, '-1p-1ww', pos_b1, wform_b1, wform_i)
        self.add_feature(features, '+1p', pos_f1)
        self.add_feature(features, '+2p', pos_f2)
        self.add_feature(features, '+1p+2p', pos_f1, pos_f2)
        #self.add_feature(features, '+1p+2p+3p', pos_f1, pos_f2, pos_f3)
        self.add_feature(features, '+1pw', pos_f1, wform_i)
        self.add_feature(features, '-1p+1p', pos_b1, pos_f1)

        self.add_feature(features, '-1w', wform_b1)
        self.add_feature(features, '-1s4', wform_b1[-4:])

        self.add_feature(features, '+1w', wform_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])

        self.add_feature(features, '-2w', wform_b2)

        self.add_feature(features, '+2w', wform_f2)
        return features

    def get_feat2_features_RU(self, i, feat, context,
                              lemma_context, pos_context, feats_context,
                              joint, val_cnt):
        if feat == None:
            feat = ''

        context       = self.START +       context + self.END
        lemma_context = self.START + lemma_context + self.END
        pos_context   = self.START +   pos_context + self.END
        feats_context = [{feat: self.START[0]}, {feat: self.START[1]}] \
                      + feats_context \
                      + [{feat: self.END[0]  }, {feat: self.END[1]  }]
        i += len(self.START)

        wform = context[i]
        wform_i  = self.normalize(wform)
        wform_b1 = self.normalize(context[i - 1])
        wform_b2 = self.normalize(context[i - 2])
        wform_f1 = self.normalize(context[i + 1])
        wform_f2 = self.normalize(context[i + 2])

        lemma_i  = lemma_context[i]
        lemma_b1 = lemma_context[i - 1]
        lemma_b2 = lemma_context[i - 2]
        lemma_f1 = lemma_context[i + 1]
        lemma_f2 = lemma_context[i + 2]

        pos_i  = pos_context[i]
        pos_b1 = pos_context[i - 1]
        pos_b2 = pos_context[i - 2]
        pos_f1 = pos_context[i + 1]
        pos_f2 = pos_context[i + 2]

        #feats_i  = feats_context[i].get(feat, '') if feat else \
        #           '|'.join('='.join((x, feats_context[i][x]))
        #                        for x in sorted(feats_context[i]))
        feats_b1 = feats_context[i - 1].get(feat, '') if feat else \
                   '|'.join('='.join((x, feats_context[i - 1][x]))
                                for x in sorted(feats_context[i - 1]))
        feats_b2 = feats_context[i - 2].get(feat, '') if feat else \
                   '|'.join('='.join((x, feats_context[i - 2][x]))
                                for x in sorted(feats_context[i - 2]))
        feats_f1 = feats_context[i + 1].get(feat, '') if feat else \
                   '|'.join('='.join((x, feats_context[i + 1][x]))
                                for x in sorted(feats_context[i + 1]))
        feats_f2 = feats_context[i + 2].get(feat, '') if feat else \
                   '|'.join('='.join((x, feats_context[i + 2][x]))
                                for x in sorted(feats_context[i + 2]))

        features = self.init_features()

        self.add_feature(features, '_')
        self.add_feature(features, 'w', wform)
        if not joint:
            self.add_feature(features, 's', self.wform_shape(wform))
        len_ = len(wform_i)
        for val in range(1, 8):
            self.add_feature(features, 'p' + str(val),
                             '' if val > len_ else wform_i[:val])
        for val in range(1, 8):
            self.add_feature(features, 's' + str(val),
                             '' if val > len_ else wform_i[-val:])
        for val in range(3, 7):
            _len = len_ - 1 - val
            for start in range(1, _len):
                self.add_feature(features, '<' + str(val) + ':' + str(start),
                                 wform_i[len_ - start - val:len_ - start]) 
        for val in range(2, 7):
            self.add_feature(features, 'S' + str(val),
                             '' if val > len_ else wform_i[val:])
        self.add_feature(features, 'p', pos_i)

        if joint or val_cnt < 7:
            self.add_feature(features, '-1f', feats_b1)
            self.add_feature(features, '-1fw', feats_b1, wform_i)
            self.add_feature(features, '-1fl', feats_b1, lemma_i)
            self.add_feature(features, '-1fp', feats_b1, pos_i)
            self.add_feature(features, '+1f', feats_f1)
            self.add_feature(features, '+1fw', feats_f1, wform_i)
            self.add_feature(features, '+1fl', feats_f1, lemma_i)
            self.add_feature(features, '+1fp', feats_f1, pos_i)
            self.add_feature(features, '-1f+1f', feats_b1, feats_f1)
        else:
            self.add_feature(features, '-1f-2f', feats_b1, feats_b2)
            self.add_feature(features, '-1fw', feats_b1, wform_i)
            self.add_feature(features, '-f',
                             feats_b1 if feats_b1 != '' else feats_b2)
            self.add_feature(features, '-fw',
                             feats_b1 if feats_b1 != '' else feats_b2, wform_i)
            self.add_feature(features, '+1f+2f', feats_f1, feats_f2)
            self.add_feature(features, '+1fw', feats_f1, wform_i)
            self.add_feature(features, '+f',
                             feats_f1 if feats_f1 != '' else feats_f2)
            self.add_feature(features, '+fw',
                             feats_f1 if feats_f1 != '' else feats_f2, wform_i)
            self.add_feature(features, '-f+f',
                             feats_b1 if feats_b1 != '' else feats_b2,
                             feats_f1 if feats_f1 != '' else feats_f2)

        self.add_feature(features, '-1l', lemma_b1)
        self.add_feature(features, '-1s4', wform_b1[-4:])
        self.add_feature(features, '-1p', pos_b1)

        self.add_feature(features, '+1l', lemma_f1)
        self.add_feature(features, '+1s3', wform_f1[-3:])
        self.add_feature(features, '+1p', pos_f1)

        self.add_feature(features, '-2l', lemma_b2)
        self.add_feature(features, '-2p', pos_b2)

        self.add_feature(features, '+2l', lemma_f2)
        self.add_feature(features, '+2p', pos_f2)
        return features



# -*- coding: utf-8 -*-
# Morra project: Base parser
#
# Copyright (C) 2020-present by Sergei Ternovykh
# License: BSD, see LICENSE for details
"""
Base classes for the project.
"""
from collections import OrderedDict, defaultdict
from copy import deepcopy
from math import isclose
import pickle
import random
from random import random as rand
from scipy.special import softmax
import sys

from corpuscula import Conllu, CorpusDict
from corpuscula.corpus_utils import _AbstractCorpus
from corpuscula.utils import LOG_FILE, print_progress


class _AveragedPerceptron:
    """This class is a port of the Textblob Averaged Perceptron Tagger
    Copyright 2013 Matthew Honnibal
    URL: <https://github.com/sloria/textblob-aptagger>

    See more implementation details here:
        https://explosion.ai/blog/part-of-speech-pos-tagger-in-python

    Small changes and extra comments where added"""

    def __init__(self, default_class=None):
        """
        :param default_class: the most common class of the train set;
                              uses only on the first train iterations
        :type default_class: str
        """
        # Each feature gets its own weight vector, so weights is a dict-of-dicts
        self.weights = {}
        # The accumulated values, for the averaging. These will be keyed by
        # feature/clas tuples
        self._totals = defaultdict(int)
        # The last time the feature was changed, for the averaging. Also
        # keyed by feature/clas tuples
        # (tstamps is short for timestamps)
        self._tstamps = defaultdict(int)
        # Number of instances seen
        self.i = 0
        # If None than you must be ready to handle the None class in the
        # beginning of training, when model has no classes at all (e.g.,
        # you can ignore None class or replace it with smth meaningful)
        self.default_class = default_class

    def predict(self, features, suggest=None, suggest_coef=None,
                with_score=False, dropout=0):
        """Dot-product the features and current weights and return the best
        label.

        :type suggest: str
        :type suggest_coef: float
        Suggestion and relevance coef of that suggestion given from some other
        tagger. May be omitted

        :param with_score: return not only most probable class but also its score
        :param dropout: a fraction of weiths to be randomly set to 0 at each
                        predict. Use it during training time to prevent
                        overfitting
        :type dropout: float
        :rtype: str | if with_score: tuple(str, float)
        """
        scores = defaultdict(float)
        for feat, value in features.items() if isinstance(features,
                                                          OrderedDict) else \
                           sorted(features.items()):
            if feat not in self.weights or value == 0:
                continue
            weights = self.weights[feat]
            for label, weight in weights.items():
                if not dropout or rand() >= dropout:
                    scores[label] += value * weight# / (1 - dropout)
        classes = scores.keys()
        if classes:
            # Increase scores of the suggested class by the relevance coef
            # of the suggestion
            if suggest:
                scores[suggest] *= 1 + suggest_coef
            # Round all scores off to avoid inaccuracy of floating point
            # operations
            scores = defaultdict(float, zip(
                classes, map(lambda x: round(x, 3), scores.values())
            ))
            # Do a secondary alphabetic sort, for stability
            res = max(classes, key=lambda label: (scores[label], label))
        else:
            res = self.default_class
        if with_score:
            scores = list(zip(*scores.items()))
            res = res, dict(zip(scores[0], softmax(scores[1])))[res] \
                           if scores else \
                       None
        return res

    def update(self, truth, guess, features):
        """Update the feature weights"""
        # self.weights: {feature: {class: weight}}
        def upd_feat(c, f, w, v):
            # param = (feature, class)
            param = (f, c)
            # для каждой пары "фича-класс" сохраняем:
            #     сумму весов на всех предыдущих итерациях
            self._totals[param] += (self.i - self._tstamps[param]) * w
            #     номер итерации, на которой вес менялся последний раз
            self._tstamps[param] = self.i
            # изменяем вес
            self.weights[f][c] = w + v

        # увеличиваем номер итерации
        self.i += 1
        if truth == guess:
            return
        # обновляем веса, только если угадали неправильно
        for f in features:
            # если весов ещё нет, но делаем нулевые веса для каждой фичи,
            # и после сразу их меняем
            weights = self.weights.setdefault(f, OrderedDict())
            # для верного класса добавляем 1
            upd_feat(truth, f, weights.get(truth, 0.0), 1.0)
            # для ошибочного вычитаем 1
            if guess is not None:
                upd_feat(guess, f, weights.get(guess, 0.0), -1.0)

    def average_weights(self):
        """Average weights from all iterations"""
        for feat, weights in self.weights.items():
            new_feat_weights = {}
            for clas, weight in weights.items():
                param = (feat, clas)
                total = self._totals[param]
                # добавляем к сумме весов актуальный вес с учётом кол-ва
                # итерация, которые он продержался
                total += (self.i - self._tstamps[param]) * weight
                # усредняем вес по всем итерациям
                averaged = round(total / self.i, 3)
                if averaged:
                    new_feat_weights[clas] = averaged
            self.weights[feat] = new_feat_weights

    def get_classes(self):
        """Classes found during train process"""
        return set(
            [x for x in self.weights.values() for x in x.keys()]
                if self.weights else
            []
        )


class BaseParser:
    """Base class for all parsers of the project"""

    def __init__(self):
        self._cdict = CorpusDict(log_file=LOG_FILE)

        self._train_corpus = []
        self._test_corpus  = []

    def backup(self):
        """Get current state"""
        return {'cdict_backup': self._cdict.backup()}

    def restore(self, o):
        """Restore current state from backup object"""
        cdict_backup = o.get('cdict_backup')
        if cdict_backup:
            self._cdict.restore(cdict_backup)

    def save(self, file_path):
        print('Saving model...', end=' ', file=LOG_FILE)
        LOG_FILE.flush()
        with open(file_path, 'wb') as f:
            pickle.dump(self.backup(), f, 2)
            print('done.', file=LOG_FILE)

    def load(self, file_path):
        print('Loading model...', end=' ', file=LOG_FILE)
        LOG_FILE.flush()
        with open(file_path, 'rb') as f:
            o = pickle.load(f)
            print('done.', file=LOG_FILE)
            self.restore(o)

    def _save_cdict(self, file_path):
        self._cdict.backup_to(file_path)

    def _load_cdict(self, file_path):
        self._cdict.restore_from(file_path)

    @staticmethod
    def load_conllu(*args, **kwargs):
        """Wrapper for ``Conllu.load()``"""
        silent = kwargs.pop('silent', None)
        if silent:
            kwargs['log_file'] = None
        elif 'log_file' not in kwargs:
            kwargs['log_file'] = LOG_FILE
        return Conllu.load(*args, **kwargs)

    @staticmethod
    def save_conllu(*args, **kwargs):
        """Wrapper for ``Conllu.save()``"""
        silent = kwargs.pop('silent', None)
        if silent:
            kwargs['log_file'] = None
        elif 'log_file' not in kwargs:
            kwargs['log_file'] = LOG_FILE
        return Conllu.save(*args, **kwargs)

    @staticmethod
    def split_corpus(corpus, split=[.8, .1, .1], save_split_to=None,
                     seed=None, silent=False):
        """Split a *corpus* in the given proportion.

        :param corpus: a name of file in CoNLL-U format or list/iterator of
                       sentences in Parsed CoNLL-U
        :param split: list of sizes of the necessary *corpus* parts. If values
                      are of int type, they are interpreted as lengths of new
                      corpora in sentences; if values are float, they are
                      proportions of a given *corpus*. The types of the
                      *split* values can't be mixed: they are either all int,
                      or all float. The sum of float values must be less or
                      equals to 1; the sum of int values can't be greater than
                      the lentgh of the *corpus*
        :param save_split_to: list of file names to save the result of the
                              *corpus* splitting. Can be `None` (default;
                              don't save parts to files) or its length must be
                              equal to the length of *split*
        :param silent: if True, suppress output
        :return: a list of new corpora
        """
        assert save_split_to is None or len(save_split_to) == len(split), \
               'ERROR: lengths of split and save_split_to must be equal'
        isfloat = len([x for x in split if isinstance(x, float)]) > 0
        if isfloat:
            assert sum(split) <= 1, \
                   "ERROR: sum of split can't be greater that 1"
        corpus = list(Conllu.load(corpus,
                                  log_file=None if silent else LOG_FILE))
        corpus_len = len(corpus)
        if isfloat:
            split = list(map(lambda x: round(corpus_len * x), split))
            diff = corpus_len - sum(split)
            if abs(diff) == 1:
                split[-1] += diff
        assert sum(split) <= corpus_len, \
               "ERROR: sum of split can't be greater that corpus length"
        random.seed(seed)
        random.shuffle(corpus)
        res = []
        pos_b = 0
        for i, sp in enumerate(split):
            pos_e = pos_b + sp
            corpus_ = corpus[pos_b:pos_e]
            pos_b = pos_e
            if save_split_to:
                Conllu.save(corpus_, save_split_to[i])
            res.append(corpus_)
        return res

    @classmethod
    def _get_corpus(cls, corpus, asis=False, silent=False):
        if isinstance(corpus, str):
            corpus = cls.load_conllu(corpus, silent=silent)
        return (s[0] if not asis and isinstance(s, tuple) else s
                   for s in corpus)

    def _predict_sents(self, sentences, predict_method, save_to=None):
        silent_save = False
        if sentences is None:
            sentences = deepcopy(self._test_corpus)
            slient_save = True
        elif isinstance(sentences, type) and issubclass(sentences,
                                                        _AbstractCorpus):
            sentences = sentences.test()
        assert sentences, 'ERROR: Sentences must not be empty'
        sentences = self._get_corpus(sentences, asis=True)
        sentences = predict_method(sentences)
        if save_to:
            self.save_conllu(sentences, save_to, silent=silent_save)
            sentences = self._get_corpus(save_to, asis=True)
        return sentences

    def parse_train_corpus(self, cnt_thresh=None, ambiguity_thresh=None):
        """Create a CorpusDict for train corpus(es) loaded. For one instance it
        may be used only once. Use ``load_train_corpus()`` with append=True
        to append one more corpus to the CorpusDict after it's created.

        :type cnt_thresh: int
        :type ambiguity_thresh: float
        Params for ``CorpusDict.fit()`` method. If omitted then default values
        will be used
        """
        assert self._train_corpus, 'ERROR: Train corpus is not loaded yet'

        print('Train:', end=' ', file=LOG_FILE)
        kwargs = {}
        if cnt_thresh is not None:
            kwargs['cnt_thresh'] = cnt_thresh
        if ambiguity_thresh is not None:
            kwargs['ambiguity_thresh'] = ambiguity_thresh
        self._cdict = CorpusDict(corpus=self._train_corpus,
                                 format='conllu_parsed',
                                 **kwargs, log_file=LOG_FILE)

    def load_train_corpus(self, corpus, append=False, parse=False, test=None,
                          seed=None):
        """Load train corpus and (possibly) create a CorpusDict for it

        :param corpus: a name of file in CoNLL-U format or list/iterator of
                       sentences in Parsed CoNLL-U
        :param append: add corpus to already loaded one(s)
        :param parse: extract corpus statistics to CorpusDict right after
                      loading
        :param test: if not None, then train corpus will be shuffled and
                     specified part of it stored as test corpus
        :type test: float
        :param seed: init value for the random number generator. Used only 
                     if test is not None
        :type seed: int
        """
        assert append or not self._train_corpus, \
               'ERROR: Train corpus is already loaded. Use append=True to ' \
               'append one more corpus'

        print('Train:', end=' ', file=LOG_FILE)
        if (isinstance(corpus, type) and issubclass(corpus, _AbstractCorpus)) \
        or isinstance(corpus, _AbstractCorpus):
            corpus = corpus.train()
        corpus = self._get_corpus(corpus)
        if test is not None:
            corpus = list(corpus)
            assert test >= 0 and test <= 1, \
                   'ERROR: A value of "test" parameter must be between 0 and 1'
            test_len = round(len(corpus) * test)
            if test_len == 0:
                print('WARNING: All the corpus given will be saved as train ' \
                      'corpus (parameter "test" is too small)', file=LOG_FILE)
            else:
                if test_len == len(corpus):
                    print('WARNING: All the corpus given will be saved as ' \
                          'test corpus (parameter "test" is too large)',
                          file=LOG_FILE)
                    self.load_test_corpus(corpus, append=append)
                    corpus = None
                else:
                    random.seed(seed)
                    random.shuffle(corpus)
                    self.load_test_corpus(corpus[:test_len], append=append)
                    corpus = corpus[test_len:]
                print('stored.', file=LOG_FILE)
        if corpus:
            if append == False:
                self._train_corpus = []
            self._train_corpus.extend(corpus)
            if parse:
                self._cdict = CorpusDict(corpus=self._train_corpus,
                                         format='conllu_parsed',
                                         log_file=LOG_FILE)

    def load_test_corpus(self, corpus, append=False):
        """Load development test corpus to validate on during training
        iterations.

        :param corpus: a name of file in CoNLL-U format or list/iterator of
                       sentences in Parsed CoNLL-U
        :param append: add corpus to already loaded one(s)
        """
        assert append or not self._test_corpus, \
               'ERROR: Test corpus is already loaded. Use append=True to ' \
               'append one more corpus'

        print('Test:', end=' ', file=LOG_FILE)
        if (isinstance(corpus, type) and issubclass(corpus, _AbstractCorpus)) \
        or isinstance(corpus, _AbstractCorpus):
            try:
                corpus = corpus.dev()
            except ValueError:
                try:
                    corpus = corpus.test()
                except ValueError:
                    raise ValueError(('ERROR: {} does not have a test part. '
                                      'Use "test" attribute of '
                                      'load_train_corpus() instead')
                                         .format(corpus.name))
        corpus = self._get_corpus(corpus)
        if append == False:
            self._test_corpus = []
        self._test_corpus.extend(corpus)

    def remove_rare_feats(self, abs_thresh=None, rel_thresh=None,
                          full_rel_thresh=None):
        """Remove feats from train and test corpora, occurence of which
        in the train corpus is less then a threshold.

        :param abs_thresh: remove features if their count in the train corpus
                           is less than this value
        :type abs_thresh: int
        :param rel_thresh: remove features if their frequency with respect to
                           total feats count of the train corpus is less than
                           this value
        :type rel_thresh: float
        :param full_rel_thresh: remove features if their frequency with respect
                                to the full count of the tokens of the train
                                corpus is less than this value
        :type full_rell_thresh: float

        *rell_thresh* and *full_rell_thresh* must be between 0 and 1"""
        corpus_len = len(self._train_corpus)
        progress_step = max(int(corpus_len / 60), 1000)
        progress_check_step = min(int(corpus_len / 100), 1000)

        print('Search rare feats...', file=LOG_FILE)
        feat_freq = {}
        token_cnt = 0
        sent_no = -1
        for sent_no, sentence in enumerate(self._train_corpus):
            if not sent_no % progress_check_step:
                print_progress(sent_no, end_value=corpus_len,
                               step=progress_step)
            for token in sentence:
                token_cnt += 1
                feats = token['FEATS']
                for feat, _ in feats.items():
                    feat_freq[feat] = feat_freq.setdefault(feat, 0) + 1
        print_progress(sent_no + 1, end_value=corpus_len, step=progress_step,
                       file=LOG_FILE)
        total_cnt = sum(feat_freq.values())
        feats_to_remove = sorted(
            feat for feat, cnt in feat_freq.items()
                if (abs_thresh and cnt < abs_thresh)
                or (rel_thresh and cnt / total_cnt < rel_thresh)
                or (full_rel_thresh and cnt / token_cnt < rull_rel_thresh)
        )

        if not feats_to_remove:
            print('Finished: Nothing to do.', file=LOG_FILE)
        else:
            print('Rare feats:', feats_to_remove, file=LOG_FILE)
            train_removes = test_removes = 0
            for i, corpus in enumerate([self._train_corpus,
                                        self._test_corpus]):
                corpus_len = len(corpus)
                progress_step = max(int(corpus_len / 60), 1000)
                progress_check_step = min(int(corpus_len / 100), 1000)

                sent_no = -1
                for sent_no, sentence in enumerate(corpus):
                    if not sent_no % progress_check_step:
                        print_progress(sent_no, end_value=corpus_len,
                                       step=progress_step)
                    for token in sentence:
                        feats = token['FEATS']
                        for feat in feats_to_remove:
                            val = feats.pop(feat, None)
                            if val:
                                if i:
                                    test_removes += 1
                                else:
                                    train_removes += 1
                print_progress(sent_no + 1, end_value=corpus_len,
                               step=progress_step, file=LOG_FILE)
            print(('Finished: {} feats removes from the train corpus, '
                   '{} from the test corpus')
                      .format(train_removes, test_removes),
                  file=LOG_FILE)

    def _train_init(self, epochs, seed, allow_empty_cdict=False):
        cdict = self._cdict
        assert self._train_corpus, 'ERROR: Train corpus is not loaded'
        if not allow_empty_cdict:
            assert not cdict.isempty(), \
                   'ERROR: Train corpus is not yet prepared'
        epochs, epochs_ = epochs if isinstance(epochs, tuple) else \
                          (epochs, 0) if epochs >= 0 else \
                          (1, epochs)
        if epochs == 0:
            epochs = 1
        if epochs_ > 0:
            epochs = -epochs_
        assert epochs_ != 0 or self._test_corpus, \
               'ERROR: epochs < 0 may be used only with a test corpus'
        corpus_len = len(self._train_corpus)
        progress_step = max(int(corpus_len / 60), 1000)
        progress_check_step = min(int(corpus_len / 100), 1000)
        random.seed(seed)
        return cdict, corpus_len, progress_step, progress_check_step, \
                                                               epochs, epochs_

    def _train_eval(self, model, epoch, epochs, epochs_,
                    best_epoch, best_score, best_weights,
                    eqs, bads, prev_score,
                    td, fd, td2, fd2, tp, fp, c, n, no_train_evals,
                    f_evaluate, f_evaluate_args):
        if td is not None:
            print('dict   : correct: {}, wrong: {}, accuracy: {}'
                      .format(td, fd, round(td / (td + fd), 4)
                                          if td + fd > 0 else 1.),
                  file=LOG_FILE)
        if td2 is not None:
            print('dict_s : correct: {}, wrong: {}, accuracy: {}'
                      .format(td2, fd2, round(td2 / (td2 + fd2), 4)
                                            if td2 + fd2 > 0 else 1.),
                  file=LOG_FILE)
        if tp is not None:
            print('predict: correct: {}, wrong: {}, accuracy: {}'
                      .format(tp, fp, round(tp / (tp + fp), 4)
                                          if tp + fp > 0 else 1.),
                  file=LOG_FILE)
        sp1 = ' ' * (len(str(c)) + len(str(n)))
        sp2 = ' ' * (8 + len(str(tp)) + len(str(fp)) - len(sp1))
        print('accuracy: train during train:' + sp2 + '{}/{}={}'
                  .format(c, n, round(c / n, 4) if n > 0 else 1.),
              file=LOG_FILE)
        if not no_train_evals:
            print(' ' * 10 + 'train after train :  ' + sp1 + sp2
                + str(round(f_evaluate(gold=self._train_corpus, silent=True,
                                       **f_evaluate_args), 4)),
                  file=LOG_FILE)
        if self._test_corpus:
            weights = deepcopy(model.weights)
            model.average_weights()
            score = f_evaluate(silent=True, **f_evaluate_args)
            iseq = isclose(score, best_score)
            if score > best_score:
                bads = 0
            elif score <= prev_score:
                bads += 1
            print('Effective test accuracy: {} {}{}'
                      .format(score, '==' if iseq else
                                     '>>' if score > best_score else
                                     '<< >' if score > prev_score else
                                     '<< =' if score == prev_score else
                                     '<< <',
                              ' ({})'.format(bads) if bads else ''),
                  file=LOG_FILE)
            if iseq:
                eqs += 1
            else:
                eqs = 0
                if score > best_score:
                    best_epoch, best_score, best_weights = \
                        epoch, score, model.weights
            if epochs_ and epoch + 1 == epochs:
                epochs = epochs_
            if epoch + 1 == epochs or (epochs < 0
                                   and bads >= -epochs):#best_epoch - epoch <= epochs):
                if eqs != epoch - best_epoch:
                    print('Search finished. Return to Epoch', best_epoch,
                          file=LOG_FILE)
                    model.weights = best_weights
                else:
                    print('Search finished', file=LOG_FILE)
                eqs = -1
            else:
                model.weights = weights
        else:
            score = prev_score
        return epoch + 1, epochs, best_epoch, best_score, best_weights, \
               eqs, bads, score

    def _train_done(self, header, model, eqs, no_train_evals,
                    f_evaluate, f_evaluate_args):
        if eqs >= 0:
            model.average_weights()
        res = None
        if not no_train_evals or self._test_corpus:
            print('Final {} {}{}{} accuracy:'
                      .format(header,
                              'train' if not no_train_evals else '',
                              '/' if not no_train_evals
                                 and self._test_corpus else '',
                              'test' if self._test_corpus else ''),
                  end=' ', file=LOG_FILE)
            if not no_train_evals:
                print(round(f_evaluate(gold=self._train_corpus, silent=True,
                                       **f_evaluate_args), 4),
                      end='', file=LOG_FILE)
            if not no_train_evals and self._test_corpus:
                print(' / ', end='', file=LOG_FILE)
            if self._test_corpus:
                res = f_evaluate(silent=True, **f_evaluate_args)
                print(round(res, 4), end='', file=LOG_FILE)
            print(file=LOG_FILE)
        return res

results = []
def autotrain(train_func, *args, silent=False,
              backup_model_func=None, restore_model_func=None,
              reload_trainset_func=None, fit_params={}, params_in_process=None,
              **kwargs):
    """
    This is a tool for models' hyperparameters selection. It made simple,
    without any multiprocessing, so it's just save time of user from cyclical
    rerunnings the training process with new parameters, but it's not speedup
    the training. Howewer, user can simultaneously run several versions of
    `autotrain()` with different hyperparameters' combinations.

    :param train_func: method to train the model.

    :param args: positional args for *train_func*. Will be passed as is.

    :param silent: if True, suppress the output.

    :param backup_model_func: method to backup the model internal state. If not
                              None, it will be used to save the state of the
                              model with the highest score, for that we don't
                              train it anew

    :param restore_model_func: method to restore the model internal state. If
                               not None, then in the end, this method will be
                               invoked with the only param: internal state 
                               of the best model. There is no use or even may
                               be cause of runtime error, to specify
                               *restore_model_func* without
                               *backup_model_func*

    :param reload_trainset_func: method to load the training corpus anew. It
                                 may become necessary if you want to achieve
                                 complete reproducibility. The training corpus
                                 is randomly shuffled before each training
                                 iteration, so you should reload it before each
                                 new training if you want to have possibility
                                 to repeat the training with the best
                                 parameters and get the model with exactly the
                                 same internal state

    :param fit_params: dict of hyperparameters' keys and lists of their values
                       among which we want to find the best. It is not very
                       flexible. We just check combinations of all the given
                       values of all the given hyperparameters.
                       E.g.:
                           fit_params={
                               'dropout': [None, .005, .01],
                               'context_dropout': [None, .05, .1]
                           }
                       produces next additional keyword args for *train_func*:
                           {'dropout': None, 'context_dropout': None}
                           {'dropout': None, 'context_dropout': .05}
                           {'dropout': None, 'context_dropout': .1}
                           {'dropout': .005, 'context_dropout': None}
                           {'dropout': .005, 'context_dropout': .05}
                           {'dropout': .005, 'context_dropout': .1}
                           {'dropout': .01, 'context_dropout': None}
                           {'dropout': .01, 'context_dropout': .05}
                           {'dropout': .01, 'context_dropout': .1}

    :param params_in_process: for internal use only. Leave it as it is

    :param kwargs: keyword args for *train_func*. Will be passed as is

    When all training have done, the tool returns
    `tuple(<best model score>, <best model params>, <best model internal state>)`.
    If *backup_model_func* is `None`, `<best model internal state>` will be 
    `None`, too

    Also, `<best model internal state>` will be loaded to your model, if
    *reload_trainset_func* is not `None`.
    """
    global results
    if not silent:
        print('AUTOTRAIN model STARTED', file=LOG_FILE)
        print('Fixed params:', kwargs, file=LOG_FILE)
        print('Fitting params:', fit_params, file=LOG_FILE)
        results = []
    _kwargs = deepcopy(kwargs)
    if not params_in_process:
        params_in_process = {}
    if fit_params:
        _fit_params = deepcopy(fit_params)
        kwarg = list(_fit_params.keys())[0]
        vals = _fit_params.pop(kwarg)
        best_res, best_kwargs, best_model = -1, None, None
        if isinstance(vals, list) or isinstance(vals, tuple):
            for val in vals:
                _kwargs[kwarg] = params_in_process[kwarg] = val
                res = autotrain(train_func, *args, silent=True,
                                backup_model_func=backup_model_func,
                                reload_trainset_func=reload_trainset_func,
                                fit_params=_fit_params,
                                params_in_process=params_in_process,
                                **_kwargs)
                if res[0] > best_res:
                    best_res, best_kwargs, best_model = res
                    best_kwargs[kwarg] = val
        elif isinstance(vals, int):
            raise 'ERROR: processing int values of fit_params ' \
                  'is not implemented yet'
        elif isinstance(vals, float):
            raise 'ERROR: processing float values of fit_params ' \
                  'is not implemented yet'
        else:
            raise 'ERROR: fit_params can contain only values of ' \
                  'list, int or float'
        res = best_res, best_kwargs, best_model
    else:
        print(file=LOG_FILE)
        print('Train params:', kwargs, file=LOG_FILE)
        if reload_trainset_func:
            reload_trainset_func()
        res = train_func(*args, **kwargs), {}, backup_model_func() \
                                                   if backup_model_func else \
                                               None
        results.append((res[0], deepcopy(params_in_process)))
    if restore_model_func:
        restore_model_func(res[2])
    if not silent:
        print('AUTOTRAIN model FINISHED', file=LOG_FILE)
        if results:
            print('Results:')
            for res_ in sorted(results, key=lambda x: (1 - x[0], str(x[1]))):
                print(res_)
            print('Optimal fitting params:', res[1], file=LOG_FILE)
        print('Train quality:', res[0], file=LOG_FILE)
        print(file=LOG_FILE)
    return res

def _evaluate(test_corpus, gold_corpus, what=None, feat=None, silent=False):
    """Score a joint accuracy of the all available tagger against the
    gold standard. Extract wforms from the gold standard text, retag them
    using all the taggers, then compute a joint accuracy score.

    :param gold_corpus: a corpus of gold sentences
    :param test_corpus: a corpus of retagged sentences from the gold corpus
    :param what: 'pos'|'lemma'|'feats': what you want to evaluate.
                 if *what* is None (default) then a joint accuracy will be
                 calculated
    :type what: str
    :param feat: name of the exact feat, tagging of that you want to evaluate.
                 Used only if *what* == 'feat'. If None (default) then tagger
                 will be evaluated for all feats
    :param silent: if True then the result will not be visualized
    :return: accuracy scores of the tags in the test corpus against the gold:
             1. by tokens: the tagging of the whole token may be either
                correct or not
             2. by tags: sum of correctly detected tags to sum of all tags that
                is non-empty in either gold or test sentences (visualized only
                if what is None or if what == 'feats' and feat is None
    :rtype: tuple(float, float)
    """
    if what:
        what = what.lower()
    assert what in [None, 'lemma', 'upos', 'feats']
    gold_corpus = list(MorphParser._get_corpus(gold_corpus))
    test_corpus = list(MorphParser._get_corpus(test_corpus))
    feat_vals = {}
    for corpus in [gold_corpus, test_corpus]:
        for sent in corpus:
            if isinstance(sent, tuple):
                sent = sent[0]
            for token in sent:
                for feat, feat_val in token['FEATS'].items():
                    feat_vals.setdefault(feat, set()).add(feat_val)
    if not silent:
        print('Evaluate', file=LOG_FILE)
    n = c = nt = ct = 0
    i = -1
    for i, gold_sent in enumerate(gold_corpus):
        if not silent and not i % 100:
            print_progress(i, end_value=None, step=1000, file=LOG_FILE)
        if isinstance(gold_sent, tuple):
            gold_sent = gold_sent[0]
        test_sent = test_corpus[i]
        if isinstance(test_sent, tuple):
            test_sent = test_sent[0]
        for j, gold_token in enumerate(gold_sent):
            test_token = test_sent[j]
            nf = cf = 0
            # LEMMA
            if what is None or what == 'lemma':
                if gold_token['FORM'] and '-' not in gold_token['ID']:
                    nf += 1
                    cf += gold_token['LEMMA'] == test_token['LEMMA']
            # UPOS
            if what is None or what == 'upos':
                if gold_token['FORM'] and '-' not in gold_token['ID']:
                    nf += 1
                    cf += gold_token['UPOS'] == test_token['UPOS']
            # FEATS
            if what is None or what == 'feats':
                if gold_token['FORM'] and gold_token['LEMMA'] \
               and gold_token['UPOS'] and '-' not in gold_token['ID']:
                    gold_feats = gold_token['FEATS']
                    test_feats = test_token['FEATS']
                    for feat_ in [feat] if what == 'feats' and feat else \
                                 feat_vals:
                        gold_feat = gold_feats.get(feat_)
                        test_feat = test_feats.get(feat_)
                        if gold_feat or test_feat:
                            nf += 1
                            if gold_feat == test_feat:
                                cf += 1
            if nf > 0:
                n += 1
                c += nf == cf
                nt += nf
                ct += cf
    res = c / n if n > 0 else 1., ct / nt if nt > 0 else 1.
    if not silent:
        if i < 0:
            print('Nothing to do!', file=LOG_FILE)
        else:
            print_progress(i + 1, end_value=0, step=1000, file=LOG_FILE)
            print('   total: {} tokens, {} tags'.format(n, nt),
                  file=LOG_FILE)
            print(' correct: {} tokens, {} tags'.format(c, ct),
                  file=LOG_FILE)
            print('   wrong: {} tokens, {} tags'.format(n - c, nt - ct),
                  file=LOG_FILE)
            print('Accuracy: {} / {}'.format(res[0], res[1]),
                  file=LOG_FILE)
    return res
